@@ -423,6 +423,11 @@ class WrappedCompositeByteBuf extends CompositeByteBuf {
         return wrapped.refCnt();
     }
 
+    @Override
+    int internalRefCnt() {
+        return wrapped.internalRefCnt();
+    }
+
     @Override
     public ByteBuf duplicate() {
         return wrapped.duplicate();
