@@ -18,7 +18,9 @@ package io.netty.buffer;
 
 import io.netty.util.concurrent.FastThreadLocal;
 import io.netty.util.concurrent.FastThreadLocalThread;
+import io.netty.util.internal.PlatformDependent;
 import io.netty.util.internal.SystemPropertyUtil;
+import org.junit.Assume;
 import org.junit.Test;
 
 import java.util.ArrayList;
@@ -33,9 +35,65 @@ import java.util.concurrent.locks.LockSupport;
 import static java.util.concurrent.TimeUnit.MILLISECONDS;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
+import static org.junit.Assert.assertNotNull;
 import static org.junit.Assert.assertTrue;
 
-public class PooledByteBufAllocatorTest {
+public class PooledByteBufAllocatorTest extends AbstractByteBufAllocatorTest<PooledByteBufAllocator> {
+
+    @Override
+    protected PooledByteBufAllocator newAllocator(boolean preferDirect) {
+        return new PooledByteBufAllocator(preferDirect);
+    }
+
+    @Override
+    protected PooledByteBufAllocator newUnpooledAllocator() {
+        return new PooledByteBufAllocator(0, 0, 8192, 1);
+    }
+
+    @Override
+    protected long expectedUsedMemory(PooledByteBufAllocator allocator, int capacity) {
+        return allocator.metric().chunkSize();
+    }
+
+    @Override
+    protected long expectedUsedMemoryAfterRelease(PooledByteBufAllocator allocator, int capacity) {
+        // This is the case as allocations will start in qInit and chunks in qInit will never be released until
+        // these are moved to q000.
+        // See https://www.bsdcan.org/2006/papers/jemalloc.pdf
+        return allocator.metric().chunkSize();
+    }
+
+    @Test
+    public void testPooledUnsafeHeapBufferAndUnsafeDirectBuffer() {
+        PooledByteBufAllocator allocator = newAllocator(true);
+        ByteBuf directBuffer = allocator.directBuffer();
+        assertInstanceOf(directBuffer,
+                PlatformDependent.hasUnsafe() ? PooledUnsafeDirectByteBuf.class : PooledDirectByteBuf.class);
+        directBuffer.release();
+
+        ByteBuf heapBuffer = allocator.heapBuffer();
+        assertInstanceOf(heapBuffer,
+                PlatformDependent.hasUnsafe() ? PooledUnsafeHeapByteBuf.class : PooledHeapByteBuf.class);
+        heapBuffer.release();
+    }
+
+    @Test
+    public void testWithoutUseCacheForAllThreads() {
+        assertFalse(Thread.currentThread() instanceof FastThreadLocalThread);
+
+        PooledByteBufAllocator pool = new PooledByteBufAllocator(
+                /*preferDirect=*/ false,
+                /*nHeapArena=*/ 1,
+                /*nDirectArena=*/ 1,
+                /*pageSize=*/8192,
+                /*maxOrder=*/ 11,
+                /*tinyCacheSize=*/ 0,
+                /*smallCacheSize=*/ 0,
+                /*normalCacheSize=*/ 0,
+                /*useCacheForAllThreads=*/ false);
+        ByteBuf buf = pool.buffer(1);
+        buf.release();
+    }
 
     @Test
     public void testArenaMetricsNoCache() {
@@ -47,6 +105,18 @@ public class PooledByteBufAllocatorTest {
         testArenaMetrics0(new PooledByteBufAllocator(true, 2, 2, 8192, 11, 1000, 1000, 1000), 100, 1, 1, 0);
     }
 
+    @Test
+    public void testArenaMetricsNoCacheAlign() {
+        Assume.assumeTrue(PooledByteBufAllocator.isDirectMemoryCacheAlignmentSupported());
+        testArenaMetrics0(new PooledByteBufAllocator(true, 2, 2, 8192, 11, 0, 0, 0, true, 64), 100, 0, 100, 100);
+    }
+
+    @Test
+    public void testArenaMetricsCacheAlign() {
+        Assume.assumeTrue(PooledByteBufAllocator.isDirectMemoryCacheAlignmentSupported());
+        testArenaMetrics0(new PooledByteBufAllocator(true, 2, 2, 8192, 11, 1000, 1000, 1000, true, 64), 100, 1, 1, 0);
+    }
+
     private static void testArenaMetrics0(
             PooledByteBufAllocator allocator, int num, int expectedActive, int expectedAlloc, int expectedDealloc) {
         for (int i = 0; i < num; i++) {
@@ -54,8 +124,8 @@ public class PooledByteBufAllocatorTest {
             assertTrue(allocator.heapBuffer().release());
         }
 
-        assertArenaMetrics(allocator.directArenas(), expectedActive, expectedAlloc, expectedDealloc);
-        assertArenaMetrics(allocator.heapArenas(), expectedActive, expectedAlloc, expectedDealloc);
+        assertArenaMetrics(allocator.metric().directArenas(), expectedActive, expectedAlloc, expectedDealloc);
+        assertArenaMetrics(allocator.metric().heapArenas(), expectedActive, expectedAlloc, expectedDealloc);
     }
 
     private static void assertArenaMetrics(
@@ -75,7 +145,7 @@ public class PooledByteBufAllocatorTest {
 
     @Test
     public void testPoolChunkListMetric() {
-        for (PoolArenaMetric arenaMetric: PooledByteBufAllocator.DEFAULT.heapArenas()) {
+        for (PoolArenaMetric arenaMetric: PooledByteBufAllocator.DEFAULT.metric().heapArenas()) {
             assertPoolChunkListMetric(arenaMetric);
         }
     }
@@ -100,7 +170,7 @@ public class PooledByteBufAllocatorTest {
         PooledByteBufAllocator allocator = new PooledByteBufAllocator(true, 1, 1, 8192, 11, 0, 0, 0);
         ByteBuf buffer = allocator.heapBuffer(500);
         try {
-            PoolArenaMetric metric = allocator.heapArenas().get(0);
+            PoolArenaMetric metric = allocator.metric().heapArenas().get(0);
             PoolSubpageMetric subpageMetric = metric.smallSubpages().get(0);
             assertEquals(1, subpageMetric.maxNumElements() - subpageMetric.numAvailable());
         } finally {
@@ -113,7 +183,7 @@ public class PooledByteBufAllocatorTest {
         PooledByteBufAllocator allocator = new PooledByteBufAllocator(true, 1, 1, 8192, 11, 0, 0, 0);
         ByteBuf buffer = allocator.heapBuffer(1);
         try {
-            PoolArenaMetric metric = allocator.heapArenas().get(0);
+            PoolArenaMetric metric = allocator.metric().heapArenas().get(0);
             PoolSubpageMetric subpageMetric = metric.tinySubpages().get(0);
             assertEquals(1, subpageMetric.maxNumElements() - subpageMetric.numAvailable());
         } finally {
@@ -121,12 +191,32 @@ public class PooledByteBufAllocatorTest {
         }
     }
 
+    @Test
+    public void testAllocNotNull() {
+        PooledByteBufAllocator allocator = new PooledByteBufAllocator(true, 1, 1, 8192, 11, 0, 0, 0);
+        // Huge allocation
+        testAllocNotNull(allocator, allocator.metric().chunkSize() + 1);
+        // Normal allocation
+        testAllocNotNull(allocator, 1024);
+        // Small allocation
+        testAllocNotNull(allocator, 512);
+        // Tiny allocation
+        testAllocNotNull(allocator, 1);
+    }
+
+    private static void testAllocNotNull(PooledByteBufAllocator allocator, int capacity) {
+        ByteBuf buffer = allocator.heapBuffer(capacity);
+        assertNotNull(buffer.alloc());
+        assertTrue(buffer.release());
+        assertNotNull(buffer.alloc());
+    }
+
     @Test
     public void testFreePoolChunk() {
         int chunkSize = 16 * 1024 * 1024;
         PooledByteBufAllocator allocator = new PooledByteBufAllocator(true, 1, 0, 8192, 11, 0, 0, 0);
         ByteBuf buffer = allocator.heapBuffer(chunkSize);
-        List<PoolArenaMetric> arenas = allocator.heapArenas();
+        List<PoolArenaMetric> arenas = allocator.metric().heapArenas();
         assertEquals(1, arenas.size());
         List<PoolChunkListMetric> lists = arenas.get(0).chunkLists();
         assertEquals(6, lists.size());
@@ -150,38 +240,65 @@ public class PooledByteBufAllocatorTest {
         assertFalse(lists.get(5).iterator().hasNext());
     }
 
-    // The ThreadDeathWatcher sleeps 1s, give it double that time.
-    @Test (timeout = 2000)
-    public void testThreadCacheDestroyedByThreadDeathWatcher() {
+    @Test (timeout = 4000)
+    public void testThreadCacheDestroyedByThreadCleaner() throws InterruptedException {
+        testThreadCacheDestroyed(false);
+    }
+
+    @Test (timeout = 4000)
+    public void testThreadCacheDestroyedAfterExitRun() throws InterruptedException {
+        testThreadCacheDestroyed(true);
+    }
+
+    private static void testThreadCacheDestroyed(boolean useRunnable) throws InterruptedException {
         int numArenas = 11;
         final PooledByteBufAllocator allocator =
             new PooledByteBufAllocator(numArenas, numArenas, 8192, 1);
 
         final AtomicBoolean threadCachesCreated = new AtomicBoolean(true);
 
-        for (int i = 0; i < numArenas; i++) {
-            new FastThreadLocalThread(new Runnable() {
-                @Override
-                public void run() {
-                    ByteBuf buf = allocator.newHeapBuffer(1024, 1024);
-                    for (int i = 0; i < buf.capacity(); i++) {
-                        buf.writeByte(0);
-                    }
-
-                    // Make sure that thread caches are actually created,
-                    // so that down below we are not testing for zero
-                    // thread caches without any of them ever having been initialized.
-                    if (allocator.numThreadLocalCaches() == 0) {
-                        threadCachesCreated.set(false);
-                    }
-
-                    buf.release();
+        final Runnable task = new Runnable() {
+            @Override
+            public void run() {
+                ByteBuf buf = allocator.newHeapBuffer(1024, 1024);
+                for (int i = 0; i < buf.capacity(); i++) {
+                    buf.writeByte(0);
                 }
-            }).start();
+
+                // Make sure that thread caches are actually created,
+                // so that down below we are not testing for zero
+                // thread caches without any of them ever having been initialized.
+                if (allocator.metric().numThreadLocalCaches() == 0) {
+                    threadCachesCreated.set(false);
+                }
+
+                buf.release();
+            }
+        };
+
+        for (int i = 0; i < numArenas; i++) {
+            final FastThreadLocalThread thread;
+            if (useRunnable) {
+                thread = new FastThreadLocalThread(task);
+                assertTrue(thread.willCleanupFastThreadLocals());
+            } else {
+                thread = new FastThreadLocalThread() {
+                    @Override
+                    public void run() {
+                        task.run();
+                    }
+                };
+                assertFalse(thread.willCleanupFastThreadLocals());
+            }
+            thread.start();
+            thread.join();
         }
 
         // Wait for the ThreadDeathWatcher to have destroyed all thread caches
-        while (allocator.numThreadLocalCaches() > 0) {
+        while (allocator.metric().numThreadLocalCaches() > 0) {
+            // Signal we want to have a GC run to ensure we can process our ThreadCleanerReference
+            System.gc();
+            System.runFinalization();
             LockSupport.parkNanos(MILLISECONDS.toNanos(100));
         }
 
@@ -195,16 +312,16 @@ public class PooledByteBufAllocatorTest {
             new PooledByteBufAllocator(numHeapArenas, 0, 8192, 1);
 
         ThreadCache tcache0 = createNewThreadCache(allocator);
-        assertEquals(1, allocator.numThreadLocalCaches());
+        assertEquals(1, allocator.metric().numThreadLocalCaches());
 
         ThreadCache tcache1 = createNewThreadCache(allocator);
-        assertEquals(2, allocator.numThreadLocalCaches());
+        assertEquals(2, allocator.metric().numThreadLocalCaches());
 
         tcache0.destroy();
-        assertEquals(1, allocator.numThreadLocalCaches());
+        assertEquals(1, allocator.metric().numThreadLocalCaches());
 
         tcache1.destroy();
-        assertEquals(0, allocator.numThreadLocalCaches());
+        assertEquals(0, allocator.metric().numThreadLocalCaches());
     }
 
     @Test(timeout = 3000)
@@ -215,36 +332,36 @@ public class PooledByteBufAllocatorTest {
 
         ThreadCache tcache0 = createNewThreadCache(allocator);
         ThreadCache tcache1 = createNewThreadCache(allocator);
-        assertEquals(2, allocator.numThreadLocalCaches());
-        assertEquals(1, allocator.heapArenas().get(0).numThreadCaches());
-        assertEquals(1, allocator.heapArenas().get(1).numThreadCaches());
-        assertEquals(1, allocator.directArenas().get(0).numThreadCaches());
-        assertEquals(1, allocator.directArenas().get(0).numThreadCaches());
+        assertEquals(2, allocator.metric().numThreadLocalCaches());
+        assertEquals(1, allocator.metric().heapArenas().get(0).numThreadCaches());
+        assertEquals(1, allocator.metric().heapArenas().get(1).numThreadCaches());
+        assertEquals(1, allocator.metric().directArenas().get(0).numThreadCaches());
+        assertEquals(1, allocator.metric().directArenas().get(0).numThreadCaches());
 
         tcache1.destroy();
 
-        assertEquals(1, allocator.numThreadLocalCaches());
-        assertEquals(1, allocator.heapArenas().get(0).numThreadCaches());
-        assertEquals(0, allocator.heapArenas().get(1).numThreadCaches());
-        assertEquals(1, allocator.directArenas().get(0).numThreadCaches());
-        assertEquals(0, allocator.directArenas().get(1).numThreadCaches());
+        assertEquals(1, allocator.metric().numThreadLocalCaches());
+        assertEquals(1, allocator.metric().heapArenas().get(0).numThreadCaches());
+        assertEquals(0, allocator.metric().heapArenas().get(1).numThreadCaches());
+        assertEquals(1, allocator.metric().directArenas().get(0).numThreadCaches());
+        assertEquals(0, allocator.metric().directArenas().get(1).numThreadCaches());
 
         ThreadCache tcache2 = createNewThreadCache(allocator);
-        assertEquals(2, allocator.numThreadLocalCaches());
-        assertEquals(1, allocator.heapArenas().get(0).numThreadCaches());
-        assertEquals(1, allocator.heapArenas().get(1).numThreadCaches());
-        assertEquals(1, allocator.directArenas().get(0).numThreadCaches());
-        assertEquals(1, allocator.directArenas().get(1).numThreadCaches());
+        assertEquals(2, allocator.metric().numThreadLocalCaches());
+        assertEquals(1, allocator.metric().heapArenas().get(0).numThreadCaches());
+        assertEquals(1, allocator.metric().heapArenas().get(1).numThreadCaches());
+        assertEquals(1, allocator.metric().directArenas().get(0).numThreadCaches());
+        assertEquals(1, allocator.metric().directArenas().get(1).numThreadCaches());
 
         tcache0.destroy();
-        assertEquals(1, allocator.numThreadLocalCaches());
+        assertEquals(1, allocator.metric().numThreadLocalCaches());
 
         tcache2.destroy();
-        assertEquals(0, allocator.numThreadLocalCaches());
-        assertEquals(0, allocator.heapArenas().get(0).numThreadCaches());
-        assertEquals(0, allocator.heapArenas().get(1).numThreadCaches());
-        assertEquals(0, allocator.directArenas().get(0).numThreadCaches());
-        assertEquals(0, allocator.directArenas().get(1).numThreadCaches());
+        assertEquals(0, allocator.metric().numThreadLocalCaches());
+        assertEquals(0, allocator.metric().heapArenas().get(0).numThreadCaches());
+        assertEquals(0, allocator.metric().heapArenas().get(1).numThreadCaches());
+        assertEquals(0, allocator.metric().directArenas().get(0).numThreadCaches());
+        assertEquals(0, allocator.metric().directArenas().get(1).numThreadCaches());
     }
 
     private static ThreadCache createNewThreadCache(final PooledByteBufAllocator allocator)
