@@ -16,8 +16,9 @@
 package io.netty.buffer;
 
 import io.netty.util.internal.PlatformDependent;
-import org.junit.Assume;
+import org.junit.Assert;
 import org.junit.BeforeClass;
+import org.junit.Test;
 
 import java.nio.ByteBuffer;
 
@@ -37,4 +38,16 @@ public class ReadOnlyUnsafeDirectByteBufferBufTest extends ReadOnlyDirectByteBuf
     protected ByteBuf buffer(ByteBuffer buffer) {
         return new ReadOnlyUnsafeDirectByteBuf(UnpooledByteBufAllocator.DEFAULT, buffer);
     }
+
+    @Test
+    @Override
+    public void testMemoryAddress() {
+        ByteBuf buf = buffer(allocate(8).asReadOnlyBuffer());
+        try {
+            Assert.assertTrue(buf.hasMemoryAddress());
+            buf.memoryAddress();
+        } finally {
+            buf.release();
+        }
+    }
 }
