@@ -26,6 +26,8 @@ import io.netty.util.CharsetUtil;
 import org.junit.Before;
 import org.junit.Test;
 
+import java.util.List;
+
 import static io.netty.buffer.Unpooled.*;
 import static org.junit.Assert.*;
 
@@ -155,7 +157,7 @@ public class HAProxyMessageDecoderTest {
     @Test(expected = HAProxyProtocolException.class)
     public void testHeaderTooLong() {
         String header = "PROXY TCP4 192.168.0.1 192.168.0.11 56324 " +
-                "00000000000000000000000000000000000000000000000000000000000000000443\r\n";
+                        "00000000000000000000000000000000000000000000000000000000000000000443\r\n";
         ch.writeInbound(copiedBuffer(header, CharsetUtil.US_ASCII));
     }
 
@@ -184,7 +186,7 @@ public class HAProxyMessageDecoderTest {
 
     @Test
     public void testTransportProtocolAndAddressFamily() {
-        final byte unkown = HAProxyProxiedProtocol.UNKNOWN.byteValue();
+        final byte unknown = HAProxyProxiedProtocol.UNKNOWN.byteValue();
         final byte tcp4 = HAProxyProxiedProtocol.TCP4.byteValue();
         final byte tcp6 = HAProxyProxiedProtocol.TCP6.byteValue();
         final byte udp4 = HAProxyProxiedProtocol.UDP4.byteValue();
@@ -192,7 +194,7 @@ public class HAProxyMessageDecoderTest {
         final byte unix_stream = HAProxyProxiedProtocol.UNIX_STREAM.byteValue();
         final byte unix_dgram = HAProxyProxiedProtocol.UNIX_DGRAM.byteValue();
 
-        assertEquals(TransportProtocol.UNSPEC, TransportProtocol.valueOf(unkown));
+        assertEquals(TransportProtocol.UNSPEC, TransportProtocol.valueOf(unknown));
         assertEquals(TransportProtocol.STREAM, TransportProtocol.valueOf(tcp4));
         assertEquals(TransportProtocol.STREAM, TransportProtocol.valueOf(tcp6));
         assertEquals(TransportProtocol.STREAM, TransportProtocol.valueOf(unix_stream));
@@ -200,7 +202,7 @@ public class HAProxyMessageDecoderTest {
         assertEquals(TransportProtocol.DGRAM, TransportProtocol.valueOf(udp6));
         assertEquals(TransportProtocol.DGRAM, TransportProtocol.valueOf(unix_dgram));
 
-        assertEquals(AddressFamily.AF_UNSPEC, AddressFamily.valueOf(unkown));
+        assertEquals(AddressFamily.AF_UNSPEC, AddressFamily.valueOf(unknown));
         assertEquals(AddressFamily.AF_IPv4, AddressFamily.valueOf(tcp4));
         assertEquals(AddressFamily.AF_IPv4, AddressFamily.valueOf(udp4));
         assertEquals(AddressFamily.AF_IPv6, AddressFamily.valueOf(tcp6));
@@ -212,16 +214,16 @@ public class HAProxyMessageDecoderTest {
     @Test
     public void testV2IPV4Decode() {
         byte[] header = new byte[28];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -267,16 +269,16 @@ public class HAProxyMessageDecoderTest {
     @Test
     public void testV2UDPDecode() {
         byte[] header = new byte[28];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -322,16 +324,16 @@ public class HAProxyMessageDecoderTest {
     @Test
     public void testv2IPV6Decode() {
         byte[] header = new byte[52];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -401,16 +403,16 @@ public class HAProxyMessageDecoderTest {
     @Test
     public void testv2UnixDecode() {
         byte[] header = new byte[232];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -479,16 +481,16 @@ public class HAProxyMessageDecoderTest {
     @Test
     public void testV2LocalProtocolDecode() {
         byte[] header = new byte[28];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -534,16 +536,16 @@ public class HAProxyMessageDecoderTest {
     @Test
     public void testV2UnknownProtocolDecode() {
         byte[] header = new byte[28];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -586,21 +588,86 @@ public class HAProxyMessageDecoderTest {
         assertFalse(ch.finish());
     }
 
+    @Test
+    public void testV2WithSslTLVs() throws Exception {
+        ch = new EmbeddedChannel(new HAProxyMessageDecoder());
+
+        final byte[] bytes = {
+                13, 10, 13, 10, 0, 13, 10, 81, 85, 73, 84, 10, 33, 17, 0, 35, 127, 0, 0, 1, 127, 0, 0, 1,
+                -55, -90, 7, 89, 32, 0, 20, 5, 0, 0, 0, 0, 33, 0, 5, 84, 76, 83, 118, 49, 34, 0, 4, 76, 69, 65, 70
+        };
+
+        int startChannels = ch.pipeline().names().size();
+        assertTrue(ch.writeInbound(copiedBuffer(bytes)));
+        Object msgObj = ch.readInbound();
+        assertEquals(startChannels - 1, ch.pipeline().names().size());
+        HAProxyMessage msg = (HAProxyMessage) msgObj;
+
+        assertEquals(HAProxyProtocolVersion.V2, msg.protocolVersion());
+        assertEquals(HAProxyCommand.PROXY, msg.command());
+        assertEquals(HAProxyProxiedProtocol.TCP4, msg.proxiedProtocol());
+        assertEquals("127.0.0.1", msg.sourceAddress());
+        assertEquals("127.0.0.1", msg.destinationAddress());
+        assertEquals(51622, msg.sourcePort());
+        assertEquals(1881, msg.destinationPort());
+        final List<HAProxyTLV> tlvs = msg.tlvs();
+
+        assertEquals(3, tlvs.size());
+        final HAProxyTLV firstTlv = tlvs.get(0);
+        assertEquals(HAProxyTLV.Type.PP2_TYPE_SSL, firstTlv.type());
+        final HAProxySSLTLV sslTlv = (HAProxySSLTLV) firstTlv;
+        assertEquals(0, sslTlv.verify());
+        assertTrue(sslTlv.isPP2ClientSSL());
+        assertTrue(sslTlv.isPP2ClientCertSess());
+        assertFalse(sslTlv.isPP2ClientCertConn());
+
+        final HAProxyTLV secondTlv = tlvs.get(1);
+
+        assertEquals(HAProxyTLV.Type.PP2_TYPE_SSL_VERSION, secondTlv.type());
+        ByteBuf secondContentBuf = secondTlv.content();
+        byte[] secondContent = new byte[secondContentBuf.readableBytes()];
+        secondContentBuf.readBytes(secondContent);
+        assertArrayEquals("TLSv1".getBytes(CharsetUtil.US_ASCII), secondContent);
+
+        final HAProxyTLV thirdTLV = tlvs.get(2);
+        assertEquals(HAProxyTLV.Type.PP2_TYPE_SSL_CN, thirdTLV.type());
+        ByteBuf thirdContentBuf = thirdTLV.content();
+        byte[] thirdContent = new byte[thirdContentBuf.readableBytes()];
+        thirdContentBuf.readBytes(thirdContent);
+        assertArrayEquals("LEAF".getBytes(CharsetUtil.US_ASCII), thirdContent);
+
+        assertTrue(sslTlv.encapsulatedTLVs().contains(secondTlv));
+        assertTrue(sslTlv.encapsulatedTLVs().contains(thirdTLV));
+
+        assertTrue(0 < firstTlv.refCnt());
+        assertTrue(0 < secondTlv.refCnt());
+        assertTrue(0 < thirdTLV.refCnt());
+        assertFalse(thirdTLV.release());
+        assertFalse(secondTlv.release());
+        assertTrue(firstTlv.release());
+        assertEquals(0, firstTlv.refCnt());
+        assertEquals(0, secondTlv.refCnt());
+        assertEquals(0, thirdTLV.refCnt());
+
+        assertNull(ch.readInbound());
+        assertFalse(ch.finish());
+    }
+
     @Test
     public void testV2WithTLV() {
         ch = new EmbeddedChannel(new HAProxyMessageDecoder(4));
 
         byte[] header = new byte[236];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -676,16 +743,16 @@ public class HAProxyMessageDecoderTest {
     @Test(expected = HAProxyProtocolException.class)
     public void testV2InvalidProtocol() {
         byte[] header = new byte[28];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -717,16 +784,16 @@ public class HAProxyMessageDecoderTest {
     @Test(expected = HAProxyProtocolException.class)
     public void testV2MissingParams() {
         byte[] header = new byte[26];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -755,16 +822,16 @@ public class HAProxyMessageDecoderTest {
     @Test(expected = HAProxyProtocolException.class)
     public void testV2InvalidCommand() {
         byte[] header = new byte[28];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -796,16 +863,16 @@ public class HAProxyMessageDecoderTest {
     @Test(expected = HAProxyProtocolException.class)
     public void testV2InvalidVersion() {
         byte[] header = new byte[28];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -839,16 +906,16 @@ public class HAProxyMessageDecoderTest {
         ch = new EmbeddedChannel(new HAProxyMessageDecoder(0));
 
         byte[] header = new byte[248];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
@@ -880,16 +947,16 @@ public class HAProxyMessageDecoderTest {
     @Test
     public void testV2IncompleteHeader() {
         byte[] header = new byte[13];
-        header[0]  = 0x0D; // Binary Prefix
-        header[1]  = 0x0A; // -----
-        header[2]  = 0x0D; // -----
-        header[3]  = 0x0A; // -----
-        header[4]  = 0x00; // -----
-        header[5]  = 0x0D; // -----
-        header[6]  = 0x0A; // -----
-        header[7]  = 0x51; // -----
-        header[8]  = 0x55; // -----
-        header[9]  = 0x49; // -----
+        header[0] = 0x0D; // Binary Prefix
+        header[1] = 0x0A; // -----
+        header[2] = 0x0D; // -----
+        header[3] = 0x0A; // -----
+        header[4] = 0x00; // -----
+        header[5] = 0x0D; // -----
+        header[6] = 0x0A; // -----
+        header[7] = 0x51; // -----
+        header[8] = 0x55; // -----
+        header[9] = 0x49; // -----
         header[10] = 0x54; // -----
         header[11] = 0x0A; // -----
 
