@@ -26,41 +26,26 @@ import io.netty.handler.codec.http.HttpResponse;
 import io.netty.handler.codec.http.HttpUtil;
 import io.netty.util.AsciiString;
 import io.netty.util.ReferenceCountUtil;
+import org.hamcrest.core.IsEqual;
 import org.junit.Test;
 
 import java.util.Arrays;
+import java.util.Collections;
+import java.util.List;
 import java.util.concurrent.Callable;
 
-import static io.netty.handler.codec.http.HttpHeaderNames.ACCESS_CONTROL_ALLOW_CREDENTIALS;
-import static io.netty.handler.codec.http.HttpHeaderNames.ACCESS_CONTROL_ALLOW_HEADERS;
-import static io.netty.handler.codec.http.HttpHeaderNames.ACCESS_CONTROL_ALLOW_METHODS;
-import static io.netty.handler.codec.http.HttpHeaderNames.ACCESS_CONTROL_ALLOW_ORIGIN;
-import static io.netty.handler.codec.http.HttpHeaderNames.ACCESS_CONTROL_EXPOSE_HEADERS;
-import static io.netty.handler.codec.http.HttpHeaderNames.ACCESS_CONTROL_REQUEST_HEADERS;
-import static io.netty.handler.codec.http.HttpHeaderNames.ACCESS_CONTROL_REQUEST_METHOD;
-import static io.netty.handler.codec.http.HttpHeaderNames.CONNECTION;
-import static io.netty.handler.codec.http.HttpHeaderNames.CONTENT_LENGTH;
-import static io.netty.handler.codec.http.HttpHeaderNames.DATE;
-import static io.netty.handler.codec.http.HttpHeaderNames.ORIGIN;
-import static io.netty.handler.codec.http.HttpHeaderNames.VARY;
-import static io.netty.handler.codec.http.HttpHeaderValues.KEEP_ALIVE;
+import static io.netty.handler.codec.http.HttpHeaderNames.*;
 import static io.netty.handler.codec.http.HttpHeaderValues.CLOSE;
+import static io.netty.handler.codec.http.HttpHeaderValues.KEEP_ALIVE;
 import static io.netty.handler.codec.http.HttpHeadersTestUtils.of;
-import static io.netty.handler.codec.http.HttpMethod.DELETE;
-import static io.netty.handler.codec.http.HttpMethod.GET;
-import static io.netty.handler.codec.http.HttpMethod.OPTIONS;
+import static io.netty.handler.codec.http.HttpMethod.*;
 import static io.netty.handler.codec.http.HttpResponseStatus.FORBIDDEN;
 import static io.netty.handler.codec.http.HttpResponseStatus.OK;
 import static io.netty.handler.codec.http.HttpVersion.HTTP_1_1;
-import static io.netty.handler.codec.http.cors.CorsConfigBuilder.forAnyOrigin;
-import static io.netty.handler.codec.http.cors.CorsConfigBuilder.forOrigin;
-import static io.netty.handler.codec.http.cors.CorsConfigBuilder.forOrigins;
-import static org.hamcrest.CoreMatchers.containsString;
-import static org.hamcrest.CoreMatchers.equalTo;
-import static org.hamcrest.CoreMatchers.is;
-import static org.hamcrest.CoreMatchers.notNullValue;
-import static org.hamcrest.CoreMatchers.nullValue;
+import static io.netty.handler.codec.http.cors.CorsConfigBuilder.*;
+import static org.hamcrest.CoreMatchers.*;
 import static org.hamcrest.MatcherAssert.assertThat;
+import static org.hamcrest.core.IsEqual.equalTo;
 
 public class CorsHandlerTest {
 
@@ -161,6 +146,15 @@ public class CorsHandlerTest {
         final HttpResponse response = preflightRequest(config, "http://localhost:8888", "content-type, xheader1");
         assertThat(response.headers().get(of("CustomHeader")), equalTo("somevalue"));
         assertThat(response.headers().get(VARY), equalTo(ORIGIN.toString()));
+        assertThat(response.headers().get(CONTENT_LENGTH), is("0"));
+    }
+
+    @Test
+    public void preflightRequestWithUnauthorizedOrigin() {
+        final String origin = "http://host";
+        final CorsConfig config = forOrigin("http://localhost").build();
+        final HttpResponse response = preflightRequest(config, origin, "xheader1");
+        assertThat(response.headers().contains(ACCESS_CONTROL_ALLOW_ORIGIN), is(false));
     }
 
     @Test
@@ -272,14 +266,15 @@ public class CorsHandlerTest {
     }
 
     @Test
-    public void simpleRequestShortCurcuit() {
+    public void simpleRequestShortCircuit() {
         final CorsConfig config = forOrigin("http://localhost:8080").shortCircuit().build();
         final HttpResponse response = simpleRequest(config, "http://localhost:7777");
         assertThat(response.status(), is(FORBIDDEN));
+        assertThat(response.headers().get(CONTENT_LENGTH), is("0"));
     }
 
     @Test
-    public void simpleRequestNoShortCurcuit() {
+    public void simpleRequestNoShortCircuit() {
         final CorsConfig config = forOrigin("http://localhost:8080").build();
         final HttpResponse response = simpleRequest(config, "http://localhost:7777");
         assertThat(response.status(), is(OK));
@@ -287,7 +282,7 @@ public class CorsHandlerTest {
     }
 
     @Test
-    public void shortCurcuitNonCorsRequest() {
+    public void shortCircuitNonCorsRequest() {
         final CorsConfig config = forOrigin("https://localhost").shortCircuit().build();
         final HttpResponse response = simpleRequest(config, null);
         assertThat(response.status(), is(OK));
@@ -295,7 +290,7 @@ public class CorsHandlerTest {
     }
 
     @Test
-    public void shortCurcuitWithConnectionKeepAliveShouldStayOpen() {
+    public void shortCircuitWithConnectionKeepAliveShouldStayOpen() {
         final CorsConfig config = forOrigin("http://localhost:8080").shortCircuit().build();
         final EmbeddedChannel channel = new EmbeddedChannel(new CorsHandler(config));
         final FullHttpRequest request = createHttpRequest(GET);
@@ -313,7 +308,7 @@ public class CorsHandlerTest {
     }
 
     @Test
-    public void shortCurcuitWithoutConnectionShouldStayOpen() {
+    public void shortCircuitWithoutConnectionShouldStayOpen() {
         final CorsConfig config = forOrigin("http://localhost:8080").shortCircuit().build();
         final EmbeddedChannel channel = new EmbeddedChannel(new CorsHandler(config));
         final FullHttpRequest request = createHttpRequest(GET);
@@ -330,7 +325,7 @@ public class CorsHandlerTest {
     }
 
     @Test
-    public void shortCurcuitWithConnectionCloseShouldClose() {
+    public void shortCircuitWithConnectionCloseShouldClose() {
         final CorsConfig config = forOrigin("http://localhost:8080").shortCircuit().build();
         final EmbeddedChannel channel = new EmbeddedChannel(new CorsHandler(config));
         final FullHttpRequest request = createHttpRequest(GET);
@@ -420,6 +415,46 @@ public class CorsHandlerTest {
         assertThat(channel.finish(), is(false));
     }
 
+    @Test
+    public void differentConfigsPerOrigin() {
+        String host1 = "http://host1:80";
+        String host2 = "http://host2";
+        CorsConfig rule1 = forOrigin(host1).allowedRequestMethods(HttpMethod.GET).build();
+        CorsConfig rule2 = forOrigin(host2).allowedRequestMethods(HttpMethod.GET, HttpMethod.POST)
+                .allowCredentials().build();
+
+        List<CorsConfig> corsConfigs = Arrays.asList(rule1, rule2);
+
+        final HttpResponse preFlightHost1 = preflightRequest(corsConfigs, host1, "", false);
+        assertThat(preFlightHost1.headers().get(ACCESS_CONTROL_ALLOW_METHODS), is("GET"));
+        assertThat(preFlightHost1.headers().getAsString(ACCESS_CONTROL_ALLOW_CREDENTIALS), is(nullValue()));
+
+        final HttpResponse preFlightHost2 = preflightRequest(corsConfigs, host2, "", false);
+        assertValues(preFlightHost2, ACCESS_CONTROL_ALLOW_METHODS.toString(), "GET", "POST");
+        assertThat(preFlightHost2.headers().getAsString(ACCESS_CONTROL_ALLOW_CREDENTIALS), IsEqual.equalTo("true"));
+    }
+
+    @Test
+    public void specificConfigPrecedenceOverGeneric() {
+        String host1 = "http://host1";
+        String host2 = "http://host2";
+
+        CorsConfig forHost1 = forOrigin(host1).allowedRequestMethods(HttpMethod.GET).maxAge(3600L).build();
+        CorsConfig allowAll = forAnyOrigin().allowedRequestMethods(HttpMethod.POST, HttpMethod.GET, HttpMethod.OPTIONS)
+                .maxAge(1800).build();
+
+        List<CorsConfig> rules = Arrays.asList(forHost1, allowAll);
+
+        final HttpResponse host1Response = preflightRequest(rules, host1, "", false);
+        assertThat(host1Response.headers().get(ACCESS_CONTROL_ALLOW_METHODS), is("GET"));
+        assertThat(host1Response.headers().getAsString(ACCESS_CONTROL_MAX_AGE), equalTo("3600"));
+
+        final HttpResponse host2Response = preflightRequest(rules, host2, "", false);
+        assertValues(host2Response, ACCESS_CONTROL_ALLOW_METHODS.toString(), "POST", "GET", "OPTIONS");
+        assertThat(host2Response.headers().getAsString(ACCESS_CONTROL_ALLOW_ORIGIN), equalTo("*"));
+        assertThat(host2Response.headers().getAsString(ACCESS_CONTROL_MAX_AGE), equalTo("1800"));
+    }
+
     private static HttpResponse simpleRequest(final CorsConfig config, final String origin) {
         return simpleRequest(config, origin, null);
     }
@@ -449,7 +484,14 @@ public class CorsHandlerTest {
     private static HttpResponse preflightRequest(final CorsConfig config,
                                                  final String origin,
                                                  final String requestHeaders) {
-        final EmbeddedChannel channel = new EmbeddedChannel(new CorsHandler(config));
+        return preflightRequest(Collections.singletonList(config), origin, requestHeaders, config.isShortCircuit());
+    }
+
+    private static HttpResponse preflightRequest(final List<CorsConfig> configs,
+                                                 final String origin,
+                                                 final String requestHeaders,
+                                                 final boolean isSHortCircuit) {
+        final EmbeddedChannel channel = new EmbeddedChannel(new CorsHandler(configs, isSHortCircuit));
         assertThat(channel.writeInbound(optionsRequest(origin, requestHeaders, null)), is(false));
         HttpResponse response = channel.readOutbound();
         assertThat(channel.finish(), is(false));
