@@ -75,23 +75,28 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
     public static final Cumulator MERGE_CUMULATOR = new Cumulator() {
         @Override
         public ByteBuf cumulate(ByteBufAllocator alloc, ByteBuf cumulation, ByteBuf in) {
-            ByteBuf buffer;
-            if (cumulation.writerIndex() > cumulation.maxCapacity() - in.readableBytes()
-                    || cumulation.refCnt() > 1) {
-                // Expand cumulation (by replace it) when either there is not more room in the buffer
-                // or if the refCnt is greater then 1 which may happen when the user use slice().retain() or
-                // duplicate().retain().
-                //
-                // See:
-                // - https://github.com/netty/netty/issues/2327
-                // - https://github.com/netty/netty/issues/1764
-                buffer = expandCumulation(alloc, cumulation, in.readableBytes());
-            } else {
-                buffer = cumulation;
+            try {
+                final ByteBuf buffer;
+                if (cumulation.writerIndex() > cumulation.maxCapacity() - in.readableBytes()
+                    || cumulation.refCnt() > 1 || cumulation.isReadOnly()) {
+                    // Expand cumulation (by replace it) when either there is not more room in the buffer
+                    // or if the refCnt is greater then 1 which may happen when the user use slice().retain() or
+                    // duplicate().retain() or if its read-only.
+                    //
+                    // See:
+                    // - https://github.com/netty/netty/issues/2327
+                    // - https://github.com/netty/netty/issues/1764
+                    buffer = expandCumulation(alloc, cumulation, in.readableBytes());
+                } else {
+                    buffer = cumulation;
+                }
+                buffer.writeBytes(in);
+                return buffer;
+            } finally {
+                // We must release in in all cases as otherwise it may produce a leak if writeBytes(...) throw
+                // for whatever release (for example because of OutOfMemoryError)
+                in.release();
             }
-            buffer.writeBytes(in);
-            in.release();
-            return buffer;
         }
     };
 
@@ -104,41 +109,62 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
         @Override
         public ByteBuf cumulate(ByteBufAllocator alloc, ByteBuf cumulation, ByteBuf in) {
             ByteBuf buffer;
-            if (cumulation.refCnt() > 1) {
-                // Expand cumulation (by replace it) when the refCnt is greater then 1 which may happen when the user
-                // use slice().retain() or duplicate().retain().
-                //
-                // See:
-                // - https://github.com/netty/netty/issues/2327
-                // - https://github.com/netty/netty/issues/1764
-                buffer = expandCumulation(alloc, cumulation, in.readableBytes());
-                buffer.writeBytes(in);
-                in.release();
-            } else {
-                CompositeByteBuf composite;
-                if (cumulation instanceof CompositeByteBuf) {
-                    composite = (CompositeByteBuf) cumulation;
+            try {
+                if (cumulation.refCnt() > 1) {
+                    // Expand cumulation (by replace it) when the refCnt is greater then 1 which may happen when the
+                    // user use slice().retain() or duplicate().retain().
+                    //
+                    // See:
+                    // - https://github.com/netty/netty/issues/2327
+                    // - https://github.com/netty/netty/issues/1764
+                    buffer = expandCumulation(alloc, cumulation, in.readableBytes());
+                    buffer.writeBytes(in);
                 } else {
-                    composite = alloc.compositeBuffer(Integer.MAX_VALUE);
-                    composite.addComponent(true, cumulation);
+                    CompositeByteBuf composite;
+                    if (cumulation instanceof CompositeByteBuf) {
+                        composite = (CompositeByteBuf) cumulation;
+                    } else {
+                        composite = alloc.compositeBuffer(Integer.MAX_VALUE);
+                        composite.addComponent(true, cumulation);
+                    }
+                    composite.addComponent(true, in);
+                    in = null;
+                    buffer = composite;
+                }
+                return buffer;
+            } finally {
+                if (in != null) {
+                    // We must release if the ownership was not transferred as otherwise it may produce a leak if
+                    // writeBytes(...) throw for whatever release (for example because of OutOfMemoryError).
+                    in.release();
                 }
-                composite.addComponent(true, in);
-                buffer = composite;
             }
-            return buffer;
         }
     };
 
+    private static final byte STATE_INIT = 0;
+    private static final byte STATE_CALLING_CHILD_DECODE = 1;
+    private static final byte STATE_HANDLER_REMOVED_PENDING = 2;
+
     ByteBuf cumulation;
     private Cumulator cumulator = MERGE_CUMULATOR;
     private boolean singleDecode;
     private boolean decodeWasNull;
     private boolean first;
+    /**
+     * A bitmask where the bits are defined as
+     * <ul>
+     *     <li>{@link #STATE_INIT}</li>
+     *     <li>{@link #STATE_CALLING_CHILD_DECODE}</li>
+     *     <li>{@link #STATE_HANDLER_REMOVED_PENDING}</li>
+     * </ul>
+     */
+    private byte decodeState = STATE_INIT;
     private int discardAfterReads = 16;
     private int numReads;
 
     protected ByteToMessageDecoder() {
-        CodecUtil.ensureNotSharable(this);
+        ensureNotSharable();
     }
 
     /**
@@ -207,6 +233,10 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
 
     @Override
     public final void handlerRemoved(ChannelHandlerContext ctx) throws Exception {
+        if (decodeState == STATE_CALLING_CHILD_DECODE) {
+            decodeState = STATE_HANDLER_REMOVED_PENDING;
+            return;
+        }
         ByteBuf buf = cumulation;
         if (buf != null) {
             // Directly set this to null so we are sure we not access it in any other method here anymore.
@@ -248,8 +278,8 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
                 callDecode(ctx, cumulation, out);
             } catch (DecoderException e) {
                 throw e;
-            } catch (Throwable t) {
-                throw new DecoderException(t);
+            } catch (Exception e) {
+                throw new DecoderException(e);
             } finally {
                 if (cumulation != null && !cumulation.isReadable()) {
                     numReads = 0;
@@ -408,7 +438,7 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
                 }
 
                 int oldInputLength = in.readableBytes();
-                decode(ctx, in, out);
+                decodeRemovalReentryProtection(ctx, in, out);
 
                 // Check if this handler was removed before continuing the loop.
                 // If it was removed, it is not safe to continue to operate on the buffer.
@@ -429,7 +459,7 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
                 if (oldInputLength == in.readableBytes()) {
                     throw new DecoderException(
                             StringUtil.simpleClassName(getClass()) +
-                            ".decode() did not read anything but decoded a message.");
+                                    ".decode() did not read anything but decoded a message.");
                 }
 
                 if (isSingleDecode()) {
@@ -438,7 +468,7 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
             }
         } catch (DecoderException e) {
             throw e;
-        } catch (Throwable cause) {
+        } catch (Exception cause) {
             throw new DecoderException(cause);
         }
     }
@@ -451,10 +481,34 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
      * @param ctx           the {@link ChannelHandlerContext} which this {@link ByteToMessageDecoder} belongs to
      * @param in            the {@link ByteBuf} from which to read data
      * @param out           the {@link List} to which decoded messages should be added
-     * @throws Exception    is thrown if an error accour
+     * @throws Exception    is thrown if an error occurs
      */
     protected abstract void decode(ChannelHandlerContext ctx, ByteBuf in, List<Object> out) throws Exception;
 
+    /**
+     * Decode the from one {@link ByteBuf} to an other. This method will be called till either the input
+     * {@link ByteBuf} has nothing to read when return from this method or till nothing was read from the input
+     * {@link ByteBuf}.
+     *
+     * @param ctx           the {@link ChannelHandlerContext} which this {@link ByteToMessageDecoder} belongs to
+     * @param in            the {@link ByteBuf} from which to read data
+     * @param out           the {@link List} to which decoded messages should be added
+     * @throws Exception    is thrown if an error occurs
+     */
+    final void decodeRemovalReentryProtection(ChannelHandlerContext ctx, ByteBuf in, List<Object> out)
+            throws Exception {
+        decodeState = STATE_CALLING_CHILD_DECODE;
+        try {
+            decode(ctx, in, out);
+        } finally {
+            boolean removePending = decodeState == STATE_HANDLER_REMOVED_PENDING;
+            decodeState = STATE_INIT;
+            if (removePending) {
+                handlerRemoved(ctx);
+            }
+        }
+    }
+
     /**
      * Is called one last time when the {@link ChannelHandlerContext} goes in-active. Which means the
      * {@link #channelInactive(ChannelHandlerContext)} was triggered.
@@ -466,7 +520,7 @@ public abstract class ByteToMessageDecoder extends ChannelInboundHandlerAdapter
         if (in.isReadable()) {
             // Only call decode() if there is something left in the buffer to decode.
             // See https://github.com/netty/netty/issues/4386
-            decode(ctx, in, out);
+            decodeRemovalReentryProtection(ctx, in, out);
         }
     }
 
