@@ -39,6 +39,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
 
     // GZIP related
     private final ByteBufChecksum crc;
+    private final boolean decompressConcatenated;
 
     private enum GzipState {
         HEADER_START,
@@ -63,7 +64,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
      * Creates a new instance with the default wrapper ({@link ZlibWrapper#ZLIB}).
      */
     public JdkZlibDecoder() {
-        this(ZlibWrapper.ZLIB, null);
+        this(ZlibWrapper.ZLIB, null, false);
     }
 
     /**
@@ -72,7 +73,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
      * supports the preset dictionary.
      */
     public JdkZlibDecoder(byte[] dictionary) {
-        this(ZlibWrapper.ZLIB, dictionary);
+        this(ZlibWrapper.ZLIB, dictionary, false);
     }
 
     /**
@@ -81,13 +82,22 @@ public class JdkZlibDecoder extends ZlibDecoder {
      * supported atm.
      */
     public JdkZlibDecoder(ZlibWrapper wrapper) {
-        this(wrapper, null);
+        this(wrapper, null, false);
     }
 
-    private JdkZlibDecoder(ZlibWrapper wrapper, byte[] dictionary) {
+    public JdkZlibDecoder(ZlibWrapper wrapper, boolean decompressConcatenated) {
+        this(wrapper, null, decompressConcatenated);
+    }
+
+    public JdkZlibDecoder(boolean decompressConcatenated) {
+        this(ZlibWrapper.GZIP, null, decompressConcatenated);
+    }
+
+    private JdkZlibDecoder(ZlibWrapper wrapper, byte[] dictionary, boolean decompressConcatenated) {
         if (wrapper == null) {
             throw new NullPointerException("wrapper");
         }
+        this.decompressConcatenated = decompressConcatenated;
         switch (wrapper) {
             case GZIP:
                 inflater = new Inflater(true);
@@ -167,25 +177,14 @@ public class JdkZlibDecoder extends ZlibDecoder {
             inflater.setInput(array);
         }
 
-        int maxOutputLength = inflater.getRemaining() << 1;
-        ByteBuf decompressed = ctx.alloc().heapBuffer(maxOutputLength);
+        ByteBuf decompressed = ctx.alloc().heapBuffer(inflater.getRemaining() << 1);
         try {
             boolean readFooter = false;
-            byte[] outArray = decompressed.array();
             while (!inflater.needsInput()) {
+                byte[] outArray = decompressed.array();
                 int writerIndex = decompressed.writerIndex();
                 int outIndex = decompressed.arrayOffset() + writerIndex;
-                int length = decompressed.writableBytes();
-
-                if (length == 0) {
-                    // completely filled the buffer allocate a new one and start to fill it
-                    out.add(decompressed);
-                    decompressed = ctx.alloc().heapBuffer(maxOutputLength);
-                    outArray = decompressed.array();
-                    continue;
-                }
-
-                int outputLength = inflater.inflate(outArray, outIndex, length);
+                int outputLength = inflater.inflate(outArray, outIndex, decompressed.writableBytes());
                 if (outputLength > 0) {
                     decompressed.writerIndex(writerIndex + outputLength);
                     if (crc != null) {
@@ -208,6 +207,8 @@ public class JdkZlibDecoder extends ZlibDecoder {
                         readFooter = true;
                     }
                     break;
+                } else {
+                    decompressed.ensureWritable(inflater.getRemaining() << 1);
                 }
             }
 
@@ -216,7 +217,13 @@ public class JdkZlibDecoder extends ZlibDecoder {
             if (readFooter) {
                 gzipState = GzipState.FOOTER_START;
                 if (readGZIPFooter(in)) {
-                    finished = true;
+                    finished = !decompressConcatenated;
+
+                    if (!finished) {
+                        inflater.reset();
+                        crc.reset();
+                        gzipState = GzipState.HEADER_START;
+                    }
                 }
             }
         } catch (DataFormatException e) {
@@ -278,6 +285,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
                 crc.update(in.readUnsignedByte()); // operating system
 
                 gzipState = GzipState.FLG_READ;
+                // fall through
             case FLG_READ:
                 if ((flags & FEXTRA) != 0) {
                     if (in.readableBytes() < 2) {
@@ -291,6 +299,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
                     xlen |= xlen1 << 8 | xlen2;
                 }
                 gzipState = GzipState.XLEN_READ;
+                // fall through
             case XLEN_READ:
                 if (xlen != -1) {
                     if (in.readableBytes() < xlen) {
@@ -300,6 +309,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
                     in.skipBytes(xlen);
                 }
                 gzipState = GzipState.SKIP_FNAME;
+                // fall through
             case SKIP_FNAME:
                 if ((flags & FNAME) != 0) {
                     if (!in.isReadable()) {
@@ -314,6 +324,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
                     } while (in.isReadable());
                 }
                 gzipState = GzipState.SKIP_COMMENT;
+                // fall through
             case SKIP_COMMENT:
                 if ((flags & FCOMMENT) != 0) {
                     if (!in.isReadable()) {
@@ -328,6 +339,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
                     } while (in.isReadable());
                 }
                 gzipState = GzipState.PROCESS_FHCRC;
+                // fall through
             case PROCESS_FHCRC:
                 if ((flags & FHCRC) != 0) {
                     if (in.readableBytes() < 4) {
@@ -337,6 +349,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
                 }
                 crc.reset();
                 gzipState = GzipState.HEADER_END;
+                // fall through
             case HEADER_END:
                 return true;
             default:
@@ -372,7 +385,7 @@ public class JdkZlibDecoder extends ZlibDecoder {
         long readCrc = crc.getValue();
         if (crcValue != readCrc) {
             throw new DecompressionException(
-                    "CRC value missmatch. Expected: " + crcValue + ", Got: " + readCrc);
+                    "CRC value mismatch. Expected: " + crcValue + ", Got: " + readCrc);
         }
     }
 
