@@ -14,8 +14,10 @@
  */
 package io.netty.handler.codec;
 
+import io.netty.util.AsciiString;
 import org.junit.Test;
 
+import java.util.ArrayList;
 import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
@@ -30,6 +32,7 @@ import static org.junit.Assert.assertNotEquals;
 import static org.junit.Assert.assertNotSame;
 import static org.junit.Assert.assertNull;
 import static org.junit.Assert.assertTrue;
+import static org.junit.Assert.fail;
 
 /**
  * Tests for {@link DefaultHeaders}.
@@ -39,11 +42,15 @@ public class DefaultHeadersTest {
     private static final class TestDefaultHeaders extends
             DefaultHeaders<CharSequence, CharSequence, TestDefaultHeaders> {
         public TestDefaultHeaders() {
-            super(CharSequenceValueConverter.INSTANCE);
+            this(CharSequenceValueConverter.INSTANCE);
+        }
+
+        public TestDefaultHeaders(ValueConverter<CharSequence> converter) {
+            super(converter);
         }
     }
 
-    private TestDefaultHeaders newInstance() {
+    private static TestDefaultHeaders newInstance() {
         return new TestDefaultHeaders();
     }
 
@@ -102,6 +109,41 @@ public class DefaultHeadersTest {
         assertTrue(values.containsAll(asList(of("value1"), of("value2"), of("value3"))));
     }
 
+    @Test
+    public void multipleValuesPerNameIterator() {
+        TestDefaultHeaders headers = newInstance();
+        headers.add(of("name"), of("value1"));
+        headers.add(of("name"), of("value2"));
+        headers.add(of("name"), of("value3"));
+        assertEquals(3, headers.size());
+
+        List<CharSequence> values = new ArrayList<CharSequence>();
+        Iterator<CharSequence> itr = headers.valueIterator(of("name"));
+        while (itr.hasNext()) {
+            values.add(itr.next());
+        }
+        assertEquals(3, values.size());
+        assertTrue(values.containsAll(asList(of("value1"), of("value2"), of("value3"))));
+    }
+
+    @Test
+    public void multipleValuesPerNameIteratorEmpty() {
+        TestDefaultHeaders headers = newInstance();
+
+        List<CharSequence> values = new ArrayList<CharSequence>();
+        Iterator<CharSequence> itr = headers.valueIterator(of("name"));
+        while (itr.hasNext()) {
+            values.add(itr.next());
+        }
+        assertEquals(0, values.size());
+        try {
+            itr.next();
+            fail();
+        } catch (NoSuchElementException ignored) {
+            // ignored
+        }
+    }
+
     @Test
     public void testContains() {
         TestDefaultHeaders headers = newInstance();
@@ -399,6 +441,24 @@ public class DefaultHeadersTest {
         assertFalse(headers.contains(of("name1"), of("value3")));
     }
 
+    @Test
+    public void testEntryEquals() {
+        Map.Entry<CharSequence, CharSequence> same1 = newInstance().add("name", "value").iterator().next();
+        Map.Entry<CharSequence, CharSequence> same2 = newInstance().add("name", "value").iterator().next();
+        assertEquals(same1, same2);
+        assertEquals(same1.hashCode(), same2.hashCode());
+
+        Map.Entry<CharSequence, CharSequence> nameDifferent1 = newInstance().add("name1", "value").iterator().next();
+        Map.Entry<CharSequence, CharSequence> nameDifferent2 = newInstance().add("name2", "value").iterator().next();
+        assertNotEquals(nameDifferent1, nameDifferent2);
+        assertNotEquals(nameDifferent1.hashCode(), nameDifferent2.hashCode());
+
+        Map.Entry<CharSequence, CharSequence> valueDifferent1 = newInstance().add("name", "value1").iterator().next();
+        Map.Entry<CharSequence, CharSequence> valueDifferent2 = newInstance().add("name", "value2").iterator().next();
+        assertNotEquals(valueDifferent1, valueDifferent2);
+        assertNotEquals(valueDifferent1.hashCode(), valueDifferent2.hashCode());
+    }
+
     @Test
     public void getAllReturnsEmptyListForUnknownName() {
         TestDefaultHeaders headers = newInstance();
@@ -472,4 +532,167 @@ public class DefaultHeadersTest {
         headers = newInstance();
         assertEquals("TestDefaultHeaders[]", headers.toString());
     }
+
+    @Test
+    public void testNotThrowWhenConvertFails() {
+        TestDefaultHeaders headers = new TestDefaultHeaders(new ValueConverter<CharSequence>() {
+            @Override
+            public CharSequence convertObject(Object value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertBoolean(boolean value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public boolean convertToBoolean(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertByte(byte value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public byte convertToByte(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertChar(char value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public char convertToChar(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertShort(short value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public short convertToShort(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertInt(int value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public int convertToInt(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertLong(long value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public long convertToLong(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertTimeMillis(long value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public long convertToTimeMillis(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertFloat(float value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public float convertToFloat(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public CharSequence convertDouble(double value) {
+                throw new IllegalArgumentException();
+            }
+
+            @Override
+            public double convertToDouble(CharSequence value) {
+                throw new IllegalArgumentException();
+            }
+        });
+        headers.set("name1", "");
+        assertNull(headers.getInt("name1"));
+        assertEquals(1, headers.getInt("name1", 1));
+
+        assertNull(headers.getBoolean(""));
+        assertFalse(headers.getBoolean("name1", false));
+
+        assertNull(headers.getByte("name1"));
+        assertEquals(1, headers.getByte("name1", (byte) 1));
+
+        assertNull(headers.getChar("name"));
+        assertEquals('n', headers.getChar("name1", 'n'));
+
+        assertNull(headers.getDouble("name"));
+        assertEquals(1, headers.getDouble("name1", 1), 0);
+
+        assertNull(headers.getFloat("name"));
+        assertEquals(Float.MAX_VALUE, headers.getFloat("name1", Float.MAX_VALUE), 0);
+
+        assertNull(headers.getLong("name"));
+        assertEquals(Long.MAX_VALUE, headers.getLong("name1", Long.MAX_VALUE));
+
+        assertNull(headers.getShort("name"));
+        assertEquals(Short.MAX_VALUE, headers.getShort("name1", Short.MAX_VALUE));
+
+        assertNull(headers.getTimeMillis("name"));
+        assertEquals(Long.MAX_VALUE, headers.getTimeMillis("name1", Long.MAX_VALUE));
+    }
+
+    @Test
+    public void testGetBooleanInvalidValue() {
+        TestDefaultHeaders headers = newInstance();
+        headers.set("name1", "invalid");
+        headers.set("name2", new AsciiString("invalid"));
+        headers.set("name3", new StringBuilder("invalid"));
+
+        assertFalse(headers.getBoolean("name1", false));
+        assertFalse(headers.getBoolean("name2", false));
+        assertFalse(headers.getBoolean("name3", false));
+    }
+
+    @Test
+    public void testGetBooleanFalseValue() {
+        TestDefaultHeaders headers = newInstance();
+        headers.set("name1", "false");
+        headers.set("name2", new AsciiString("false"));
+        headers.set("name3", new StringBuilder("false"));
+
+        assertFalse(headers.getBoolean("name1", true));
+        assertFalse(headers.getBoolean("name2", true));
+        assertFalse(headers.getBoolean("name3", true));
+    }
+
+    @Test
+    public void testGetBooleanTrueValue() {
+        TestDefaultHeaders headers = newInstance();
+        headers.set("name1", "true");
+        headers.set("name2", new AsciiString("true"));
+        headers.set("name3", new StringBuilder("true"));
+
+        assertTrue(headers.getBoolean("name1", false));
+        assertTrue(headers.getBoolean("name2", false));
+        assertTrue(headers.getBoolean("name3", false));
+    }
 }
