@@ -15,7 +15,6 @@
  */
 package io.netty.util;
 
-import io.netty.util.ByteProcessor.IndexOfProcessor;
 import io.netty.util.internal.EmptyArrays;
 import io.netty.util.internal.InternalThreadLocalMap;
 import io.netty.util.internal.PlatformDependent;
@@ -37,7 +36,7 @@ import static io.netty.util.internal.ObjectUtil.checkNotNull;
  * A string which has been encoded into a character encoding whose character always takes a single byte, similarly to
  * ASCII. It internally keeps its content in a byte array unlike {@link String}, which uses a character array, for
  * reduced memory footprint and faster data transfer from/to byte-based data structures such as a byte array and
- * {@link ByteBuffer}. It is often used in conjunction with {@link Headers} that require a {@link CharSequence}.
+ * {@link ByteBuffer}. It is often used in conjunction with {@code Headers} that require a {@link CharSequence}.
  * <p>
  * This class was designed to provide an immutable array of bytes, and caches some internal state based upon the value
  * of this array. However underlying access to this byte array is provided via not copying the array on construction or
@@ -45,7 +44,7 @@ import static io.netty.util.internal.ObjectUtil.checkNotNull;
  * {@link #arrayChanged()} so the state of this class can be reset.
  */
 public final class AsciiString implements CharSequence, Comparable<CharSequence> {
-    public static final AsciiString EMPTY_STRING = new AsciiString("");
+    public static final AsciiString EMPTY_STRING = cached("");
     private static final char MAX_CHAR_VALUE = 255;
 
     public static final int INDEX_NOT_FOUND = -1;
@@ -147,7 +146,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
                 this.offset = start;
             }
         } else {
-            this.value = new byte[length];
+            this.value = PlatformDependent.allocateUninitializedArray(length);
             int oldPos = value.position();
             value.get(this.value, 0, length);
             value.position(oldPos);
@@ -173,7 +172,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
                             + ") <= " + "value.length(" + value.length + ')');
         }
 
-        this.value = new byte[length];
+        this.value = PlatformDependent.allocateUninitializedArray(length);
         for (int i = 0, j = start; i < length; i++, j++) {
             this.value[i] = c2b(value[j]);
         }
@@ -220,7 +219,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
                             + ") <= " + "value.length(" + value.length() + ')');
         }
 
-        this.value = new byte[length];
+        this.value = PlatformDependent.allocateUninitializedArray(length);
         for (int i = 0, j = start; i < length; i++, j++) {
             this.value[i] = c2b(value.charAt(j));
         }
@@ -484,7 +483,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
                 return that;
             }
 
-            byte[] newValue = new byte[thisLen + thatLen];
+            byte[] newValue = PlatformDependent.allocateUninitializedArray(thisLen + thatLen);
             System.arraycopy(value, arrayOffset(), newValue, 0, thisLen);
             System.arraycopy(that.value, that.arrayOffset(), newValue, thisLen, thatLen);
             return new AsciiString(newValue, false);
@@ -494,7 +493,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
             return new AsciiString(string);
         }
 
-        byte[] newValue = new byte[thisLen + thatLen];
+        byte[] newValue = PlatformDependent.allocateUninitializedArray(thisLen + thatLen);
         System.arraycopy(value, arrayOffset(), newValue, 0, thisLen);
         for (int i = thisLen, j = 0; i < newValue.length; i++, j++) {
             newValue[i] = c2b(string.charAt(j));
@@ -673,44 +672,35 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
      * @throws NullPointerException if {@code subString} is {@code null}.
      */
     public int indexOf(CharSequence subString, int start) {
+        final int subCount = subString.length();
         if (start < 0) {
             start = 0;
         }
-
-        final int thisLen = length();
-
-        int subCount = subString.length();
         if (subCount <= 0) {
-            return start < thisLen ? start : thisLen;
+            return start < length ? start : length;
         }
-        if (subCount > thisLen - start) {
-            return -1;
+        if (subCount > length - start) {
+            return INDEX_NOT_FOUND;
         }
 
         final char firstChar = subString.charAt(0);
         if (firstChar > MAX_CHAR_VALUE) {
-            return -1;
+            return INDEX_NOT_FOUND;
         }
-        ByteProcessor IndexOfVisitor = new IndexOfProcessor((byte) firstChar);
-        try {
-            for (;;) {
-                int i = forEachByte(start, thisLen - start, IndexOfVisitor);
-                if (i == -1 || subCount + i > thisLen) {
-                    return -1; // handles subCount > count || start >= count
-                }
+        final byte firstCharAsByte = c2b0(firstChar);
+        final int len = offset + length - subCount;
+        for (int i = start + offset; i <= len; ++i) {
+            if (value[i] == firstCharAsByte) {
                 int o1 = i, o2 = 0;
-                while (++o2 < subCount && b2c(value[++o1 + arrayOffset()]) == subString.charAt(o2)) {
+                while (++o2 < subCount && b2c(value[++o1]) == subString.charAt(o2)) {
                     // Intentionally empty
                 }
                 if (o2 == subCount) {
-                    return i;
+                    return i - offset;
                 }
-                start = i + 1;
             }
-        } catch (Exception e) {
-            PlatformDependent.throwException(e);
-            return -1;
         }
+        return INDEX_NOT_FOUND;
     }
 
     /**
@@ -723,22 +713,22 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
      * -1 if found no occurrence.
      */
     public int indexOf(char ch, int start) {
+        if (ch > MAX_CHAR_VALUE) {
+            return INDEX_NOT_FOUND;
+        }
+
         if (start < 0) {
             start = 0;
         }
 
-        final int thisLen = length();
-
-        if (ch > MAX_CHAR_VALUE) {
-            return -1;
-        }
-
-        try {
-            return forEachByte(start, thisLen - start, new IndexOfProcessor((byte) ch));
-        } catch (Exception e) {
-            PlatformDependent.throwException(e);
-            return -1;
+        final byte chAsByte = c2b0(ch);
+        final int len = offset + length;
+        for (int i = start + offset; i < len; ++i) {
+            if (value[i] == chAsByte) {
+                return i - offset;
+            }
         }
+        return INDEX_NOT_FOUND;
     }
 
     /**
@@ -766,44 +756,35 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
      * @throws NullPointerException if {@code subString} is {@code null}.
      */
     public int lastIndexOf(CharSequence subString, int start) {
-        final int thisLen = length();
         final int subCount = subString.length();
-
-        if (subCount > thisLen || start < 0) {
-            return -1;
+        if (start < 0) {
+            start = 0;
         }
-
         if (subCount <= 0) {
-            return start < thisLen ? start : thisLen;
+            return start < length ? start : length;
+        }
+        if (subCount > length - start) {
+            return INDEX_NOT_FOUND;
         }
 
-        start = Math.min(start, thisLen - subCount);
-
-        // count and subCount are both >= 1
         final char firstChar = subString.charAt(0);
         if (firstChar > MAX_CHAR_VALUE) {
-            return -1;
+            return INDEX_NOT_FOUND;
         }
-        ByteProcessor IndexOfVisitor = new IndexOfProcessor((byte) firstChar);
-        try {
-            for (;;) {
-                int i = forEachByteDesc(start, thisLen - start, IndexOfVisitor);
-                if (i == -1) {
-                    return -1;
-                }
+        final byte firstCharAsByte = c2b0(firstChar);
+        final int end = offset + start;
+        for (int i = offset + length - subCount; i >= end; --i) {
+            if (value[i] == firstCharAsByte) {
                 int o1 = i, o2 = 0;
-                while (++o2 < subCount && b2c(value[++o1 + arrayOffset()]) == subString.charAt(o2)) {
+                while (++o2 < subCount && b2c(value[++o1]) == subString.charAt(o2)) {
                     // Intentionally empty
                 }
                 if (o2 == subCount) {
-                    return i;
+                    return i - offset;
                 }
-                start = i - 1;
             }
-        } catch (Exception e) {
-            PlatformDependent.throwException(e);
-            return -1;
         }
+        return INDEX_NOT_FOUND;
     }
 
     /**
@@ -895,29 +876,23 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
             return this;
         }
 
-        final int index;
-        final byte oldCharByte = c2b(oldChar);
-        try {
-            index = forEachByte(new IndexOfProcessor(oldCharByte));
-        } catch (Exception e) {
-            PlatformDependent.throwException(e);
-            return this;
-        }
-        if (index == -1) {
-            return this;
-        }
-
-        final byte newCharByte = c2b(newChar);
-        byte[] buffer = new byte[length()];
-        for (int i = 0, j = arrayOffset(); i < buffer.length; i++, j++) {
-            byte b = value[j];
-            if (b == oldCharByte) {
-                b = newCharByte;
+        final byte oldCharAsByte = c2b0(oldChar);
+        final byte newCharAsByte = c2b(newChar);
+        final int len = offset + length;
+        for (int i = offset; i < len; ++i) {
+            if (value[i] == oldCharAsByte) {
+                byte[] buffer = PlatformDependent.allocateUninitializedArray(length());
+                System.arraycopy(value, offset, buffer, 0, i - offset);
+                buffer[i - offset] = newCharAsByte;
+                ++i;
+                for (; i < len; ++i) {
+                    byte oldValue = value[i];
+                    buffer[i - offset] = oldValue != oldCharAsByte ? oldValue : newCharAsByte;
+                }
+                return new AsciiString(buffer, false);
             }
-            buffer[i] = b;
         }
-
-        return new AsciiString(buffer, false);
+        return this;
     }
 
     /**
@@ -967,7 +942,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
             return this;
         }
 
-        final byte[] newValue = new byte[length()];
+        final byte[] newValue = PlatformDependent.allocateUninitializedArray(length());
         for (i = 0, j = arrayOffset(); i < newValue.length; ++i, ++j) {
             newValue[i] = toLowerCase(value[j]);
         }
@@ -997,7 +972,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
             return this;
         }
 
-        final byte[] newValue = new byte[length()];
+        final byte[] newValue = PlatformDependent.allocateUninitializedArray(length());
         for (i = 0, j = arrayOffset(); i < newValue.length; ++i, ++j) {
             newValue[i] = toUpperCase(value[j]);
         }
@@ -1006,7 +981,36 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
     }
 
     /**
-     * Copies this string removing white space characters from the beginning and end of the string.
+     * Copies this string removing white space characters from the beginning and end of the string, and tries not to
+     * copy if possible.
+     *
+     * @param c The {@link CharSequence} to trim.
+     * @return a new string with characters {@code <= \\u0020} removed from the beginning and the end.
+     */
+    public static CharSequence trim(CharSequence c) {
+        if (c.getClass() == AsciiString.class) {
+            return ((AsciiString) c).trim();
+        }
+        if (c instanceof String) {
+            return ((String) c).trim();
+        }
+        int start = 0, last = c.length() - 1;
+        int end = last;
+        while (start <= end && c.charAt(start) <= ' ') {
+            start++;
+        }
+        while (end >= start && c.charAt(end) <= ' ') {
+            end--;
+        }
+        if (start == 0 && end == last) {
+            return c;
+        }
+        return c.subSequence(start, end);
+    }
+
+    /**
+     * Duplicates this string removing white space characters from the beginning and end of the
+     * string, without copying.
      *
      * @return a new string with characters {@code <= \\u0020} removed from the beginning and the end.
      */
@@ -1111,7 +1115,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
             }
         }
 
-        return res.toArray(new AsciiString[res.size()]);
+        return res.toArray(new AsciiString[0]);
     }
 
     /**
@@ -1121,10 +1125,12 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
      */
     @Override
     public int hashCode() {
-        if (hash == 0) {
-            hash = PlatformDependent.hashCodeAscii(value, offset, length);
+        int h = hash;
+        if (h == 0) {
+            h = PlatformDependent.hashCodeAscii(value, offset, length);
+            hash = h;
         }
-        return hash;
+        return h;
     }
 
     @Override
@@ -1144,20 +1150,21 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
 
     /**
      * Translates the entire byte string to a {@link String}.
-     * @see {@link #toString(int)}
+     * @see #toString(int)
      */
     @Override
     public String toString() {
-        if (string != null) {
-            return string;
+        String cache = string;
+        if (cache == null) {
+            cache = toString(0);
+            string = cache;
         }
-        string = toString(0);
-        return string;
+        return cache;
     }
 
     /**
      * Translates the entire byte string to a {@link String} using the {@code charset} encoding.
-     * @see {@link #toString(int, int)}
+     * @see #toString(int, int)
      */
     public String toString(int start) {
         return toString(start, length());
@@ -1384,6 +1391,18 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
         return string.getClass() == AsciiString.class ? (AsciiString) string : new AsciiString(string);
     }
 
+    /**
+     * Returns an {@link AsciiString} containing the given string and retains/caches the input
+     * string for later use in {@link #toString()}.
+     * Used for the constants (which already stored in the JVM's string table) and in cases
+     * where the guaranteed use of the {@link #toString()} method.
+     */
+    public static AsciiString cached(String string) {
+        AsciiString asciiString = new AsciiString(string);
+        asciiString.string = string;
+        return asciiString;
+    }
+
     /**
      * Returns the case-insensitive hash code of the specified string. Note that this method uses the same hashing
      * algorithm with {@link #hashCode()} so that you can put both {@link AsciiString}s and arbitrary
@@ -1433,8 +1452,8 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
         if (a.length() != b.length()) {
             return false;
         }
-        for (int i = 0, j = 0; i < a.length(); ++i, ++j) {
-            if (!equalsIgnoreCase(a.charAt(i),  b.charAt(j))) {
+        for (int i = 0; i < a.length(); ++i) {
+            if (!equalsIgnoreCase(a.charAt(i),  b.charAt(i))) {
                 return false;
             }
         }
@@ -1788,10 +1807,7 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
             return INDEX_NOT_FOUND;
         }
         final int sz = cs.length();
-        if (start < 0) {
-            start = 0;
-        }
-        for (int i = start; i < sz; i++) {
+        for (int i = start < 0 ? 0 : start; i < sz; i++) {
             if (cs.charAt(i) == searchChar) {
                 return i;
             }
@@ -1835,6 +1851,10 @@ public final class AsciiString implements CharSequence, Comparable<CharSequence>
         return (byte) ((c > MAX_CHAR_VALUE) ? '?' : c);
     }
 
+    private static byte c2b0(char c) {
+        return (byte) c;
+    }
+
     public static char b2c(byte b) {
         return (char) (b & 0xFF);
     }
