@@ -18,8 +18,6 @@ package io.netty.channel.epoll;
 import io.netty.channel.Channel;
 import io.netty.channel.EventLoop;
 import io.netty.channel.socket.ServerSocketChannel;
-import io.netty.channel.unix.FileDescriptor;
-import io.netty.channel.unix.Socket;
 
 import java.io.IOException;
 import java.net.InetAddress;
@@ -29,8 +27,8 @@ import java.util.Collection;
 import java.util.Collections;
 import java.util.Map;
 
+import static io.netty.channel.epoll.LinuxSocket.newSocketStream;
 import static io.netty.channel.unix.NativeInetAddress.address;
-import static io.netty.channel.unix.Socket.newSocketStream;
 
 /**
  * {@link ServerSocketChannel} implementation that uses linux EPOLL Edge-Triggered Mode for
@@ -39,7 +37,6 @@ import static io.netty.channel.unix.Socket.newSocketStream;
 public final class EpollServerSocketChannel extends AbstractEpollServerChannel implements ServerSocketChannel {
 
     private final EpollServerSocketChannelConfig config;
-    private volatile InetSocketAddress local;
     private volatile Collection<InetAddress> tcpMd5SigAddresses = Collections.emptyList();
 
     public EpollServerSocketChannel() {
@@ -47,35 +44,19 @@ public final class EpollServerSocketChannel extends AbstractEpollServerChannel i
         config = new EpollServerSocketChannelConfig(this);
     }
 
-    /**
-     * @deprecated Use {@link #EpollServerSocketChannel(Socket, boolean)}.
-     * Creates a new {@link EpollServerSocketChannel} from an existing {@link FileDescriptor}.
-     */
-    @Deprecated
-    public EpollServerSocketChannel(FileDescriptor fd) {
+    public EpollServerSocketChannel(int fd) {
         // Must call this constructor to ensure this object's local address is configured correctly.
         // The local address can only be obtained from a Socket object.
-        this(new Socket(fd.intValue()));
+        this(new LinuxSocket(fd));
     }
 
-    /**
-     * @deprecated Use {@link #EpollServerSocketChannel(Socket, boolean)}.
-     * Creates a new {@link EpollServerSocketChannel} from an existing {@link Socket}.
-     */
-    @Deprecated
-    public EpollServerSocketChannel(Socket fd) {
+    EpollServerSocketChannel(LinuxSocket fd) {
         super(fd);
-        // As we create an EpollServerSocketChannel from a FileDescriptor we should try to obtain the remote and local
-        // address from it. This is needed as the FileDescriptor may be bound already.
-        local = fd.localAddress();
         config = new EpollServerSocketChannelConfig(this);
     }
 
-    public EpollServerSocketChannel(Socket fd, boolean active) {
+    EpollServerSocketChannel(LinuxSocket fd, boolean active) {
         super(fd, active);
-        // As we create an EpollServerSocketChannel from a FileDescriptor we should try to obtain the remote and local
-        // address from it. This is needed as the FileDescriptor may be bound already.
-        local = fd.localAddress();
         config = new EpollServerSocketChannelConfig(this);
     }
 
@@ -86,14 +67,11 @@ public final class EpollServerSocketChannel extends AbstractEpollServerChannel i
 
     @Override
     protected void doBind(SocketAddress localAddress) throws Exception {
-        InetSocketAddress addr = (InetSocketAddress) localAddress;
-        checkResolvable(addr);
-        fd().bind(addr);
-        local = fd().localAddress();
+        super.doBind(localAddress);
         if (Native.IS_SUPPORTING_TCP_FASTOPEN && config.getTcpFastopen() > 0) {
-            Native.setTcpFastopen(fd().intValue(), config.getTcpFastopen());
+            socket.setTcpFastOpen(config.getTcpFastopen());
         }
-        fd().listen(config.getBacklog());
+        socket.listen(config.getBacklog());
         active = true;
     }
 
@@ -112,14 +90,9 @@ public final class EpollServerSocketChannel extends AbstractEpollServerChannel i
         return config;
     }
 
-    @Override
-    protected InetSocketAddress localAddress0() {
-        return local;
-    }
-
     @Override
     protected Channel newChildChannel(int fd, byte[] address, int offset, int len) throws Exception {
-        return new EpollSocketChannel(this, new Socket(fd), address(address, offset, len));
+        return new EpollSocketChannel(this, new LinuxSocket(fd), address(address, offset, len));
     }
 
     Collection<InetAddress> tcpMd5SigAddresses() {
